// arch/x86/thread.cc -- Thread switching
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.

#include <kern/thread.h>
#include <kern/mem.h>
#include <kern/pagetable.h>

namespace Arch {
	void set_aspace(Mem::ProcAddrSpace *aspace)
	{
		u32 cr3 = Mem::kvirt_to_phys(aspace->page_table->toplevel);
		asm volatile("movl %0, %%cr3" : : "r" (cr3) : "memory");
	}

	void switch_thread(Threads::Thread *dest, Threads::Thread *src)
	{
		u32 dummy1, dummy2;
		
		if (dest->addr_space) {
			assert(dest->addr_space == dest->active_addr_space);
		
			if (dest->addr_space != src->active_addr_space)
				set_aspace(dest->addr_space);
		} else {
			dest->active_addr_space = src->active_addr_space;
		}
		
		Priv::tss.esp0 = reinterpret_cast<u32>(dest);

		asm volatile("movl %%esp, (%0);"
		             "movl %%ebp, 4(%0);"
		             "movb $0, 8(%0);"
		             "movb 8(%1), %%al;"
		             "cmpb $0, %%al;"
		             "movl (%1), %%esp;"
		             "movl 4(%1), %%ebp;"
		             "jnz x86_new_thread;" :
		             "=a" (dummy1), "=c" (dummy2) :
		             "0" (&src->arch.esp), "1" (&dest->arch.esp) :
		             "ebx", "edx", "esi", "edi", "memory");
	}
}
