#ifndef _KERN_ORB_H
#define _KERN_ORB_H

#include <kern/types.h>
#include <orb.h>

#include <util/list.h>
#include <util/rbtree.h>
#include <System/Objects.h>

namespace Mem {
	class ProcAddrSpace;
};

namespace Threads {
	class Thread;
};

namespace ORB {
	typedef u32 ID;

	struct CallFrame {
		// Address Space and PC to return to
		Mem::ProcAddrSpace *ret_aspace;
		ulong ret_pc;
		
		// Caller's PIB Pointer
		System::RunTime::ParamInfoBlock *caller_user_pib;
		
		// Object and Method that were called -- it probably isn't strictly
		// necessary to keep track of this here, but it'd help in doing a
		// "traceforward" of the method invocation stack in order to
		// debug a stalled method call.
		
		ID object, method;
	};
	
	struct CallStackHeader {
		Threads::Thread *thread;
		Util::List node;
		
		// Number of CallFrames under this header.
		// For a full page, this is:
		//   (Arch::page_size - sizeof(CallStackHeader)) / sizeof(CallFrame)
		// There is also a much smaller group of call frames in the Thread
		// structure, so that an ORB stack page doesn't need to be allocated
		// in the common case of shallow method nesting.
	
		int num_frames;
		
		CallFrame frames[0];
	};
	
	struct ObjectHdr;
	struct Object;
	
	typedef Util::RBTree<ObjectHdr, Object *, Object *> IDRMap;

	struct ObjectHdr {
		IDRMap::Node rbtree_node;
		u32 id;

		union {
			struct {
				u32 Pointer:1;
			};
			
			u32 flags;
		};
	};
	
	struct Object : public ObjectHdr {
		Mem::ProcAddrSpace *aspace;
		uintptr_t entry;
	};

	struct ObjectPtr : public ObjectHdr {
		Object *object;
	};

	class IDSpace {
		// Reverse mapping of object pointers to local IDs
		IDRMap id_rmap;
	
	public:
		Object *lookup(u32 id);
		u32 rlookup(Object *obj);
	};
}

#endif
