// orb/invoke.cc -- Method Invocation
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.

#include <kern/types.h>
#include <kern/orb.h>
#include <kern/pagealloc.h>
#include <kern/compiler.h>
#include <arch/usercopy.h>
#include <orb.h>

using System::RunTime::ParamInfoBlock;
using namespace ORB;
using Threads::Thread;

namespace ORB {
	static CallFrame *new_frame(Thread *thread)
	{
		CallStackHeader *hdr = thread->orbstack_top_hdr;
		
		if (unlikely(thread->orbstack_top == hdr->num_frames - 1)) {
			hdr = static_cast<CallStackHeader *>(Mem::alloc_pages(1));
			new(hdr) CallStackHeader;
			hdr->thread = thread;
			hdr->num_frames = (Arch::page_size - sizeof(CallStackHeader)) /
			                  sizeof(CallFrame);

			thread->orbstack_top_hdr->node.add_front(&hdr->node);
			thread->orbstack_top_hdr = hdr;
			thread->orbstack_top = 0;
			return &hdr->frames[0];
		}
		
		return &hdr->frames[++thread->orbstack_top];
	}
	
	u32 IDSpace::rlookup(Object *obj)
	{
#if 0
		ObjectHdr *hdr = idtree.find(obj);
		
		if (!hdr)
			return 0;
		
		return hdr->id;
#endif
		return 0;
	}
}

extern "C" void invoke_method(u32 objid, u32 methid, ParamInfoBlock *user_pib,
                              ulong ret_pc)
{
	ParamInfoBlock pib = Arch::copyin(user_pib);
	CallFrame *frame = new_frame(curthread);
	
	frame->object = objid;
	frame->method = methid;
	frame->caller_user_pib = user_pib;
	frame->ret_pc = ret_pc;
	
	printf("invoke_method: frame %p object %x method %x pib %p ret %lx\n",
	       frame, frame->object, frame->method, frame->caller_user_pib,
	       frame->ret_pc);
	
	
}
