// tests/aspace.cc
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.

#include <kern/types.h>
#include <kern/libc.h>
#include <kern/time.h>
#include <kern/thread.h>
#include <kern/mem.h>

using namespace Mem;
namespace Mem {
	extern IMappable physmem;
}

void thread(void *arg)
{
	// Yuck -- out param->return value promotion would turn it into
	// AddrSpace aspace = AddrSpace::downcast(Mem::proc_addr_space_factory.create()),
	// but that's still ugly.  Perhaps something like:
	// AddrSpace aspace(Mem::proc_addr_space_factory).
	
	Object obj;
	proc_addr_space_factory.create(&obj);
	IAddrSpace aspace = IAddrSpace::downcast(obj);
	
	printf("aspace %p created\n", (void *)aspace);
	
	ProcAddrSpace *kaspace = static_cast<ProcAddrSpace *>(AddrSpace::classptr(aspace));
	
	printf("kaspace %p\n", kaspace);
	
	curthread->set_aspace(kaspace);
	printf("Addr space set\n");

	IAddrSpace stacked[2];
	IMappable mappables[2];

	for (int i = 0; i < 2; i++) {
		addr_space_factory.create(&obj);
		stacked[i] = IAddrSpace::downcast(obj);
		stacked[i].get_mappable(&mappables[i]);
	}
	
	printf("stacked %p %p mappable %p %p\n", 
	       (void *)stacked[0], (void *)stacked[1],
	       (void *)mappables[0], (void *)mappables[1]);
		
	char *buf = new char[100000];
	printf("buf %p\n", buf);
	
	char *abuf = (char *)((((ulong)buf) + 4095) & ~4095);
	printf("abuf %p\n", abuf);
	
	memset(abuf, 'A', Arch::page_size);
	memset(abuf + Arch::page_size, 'B', Arch::page_size);

	Region region;
	u64 vstart[6];
	System::Mem::MapFlags mf;
	mf.access_IDLNS_Read = 1;
	mf.access_IDLNS_Write = 1;
	mf.Fixed = 1;
	
	region.start = 0;
	region.end = 0xffffffff;
	vstart[4] = 0xdeadbeef00000000ULL;
	vstart[5] = 0xdeadb12345678000ULL;
	
	stacked[0].map(Mem::physmem, region, &vstart[4], mf);
	mf.CopyOnWrite = 1;
	stacked[0].map(Mem::physmem, region, &vstart[5], mf);
	mf.CopyOnWrite = 0;
	printf("stacked[0] vstart %llx, %llx\n", vstart[4], vstart[5]);
	
	region.start = 0xdeadb00000000000ULL;
	region.end   = 0xdeadbfffffffffffULL;
	vstart[4]    = 0x1234500000000000ULL;
	mf.Fixed = 0;
//	mf.access_IDLNS_Write = 0;
	
	stacked[1].map(mappables[0], region, &vstart[4], mf);

	printf("stacked[1] vstart %llx\n", vstart[4]);
	
	region.start = vstart[4] + 0xeef00000000ULL + kvirt_to_phys(abuf);
	region.end   = vstart[4] + 0xeef00000000ULL + kvirt_to_phys(abuf) + Arch::page_size - 1;
	vstart[0] = System::Mem::AddrSpace_ns::unspecified_start;
	vstart[1] = System::Mem::AddrSpace_ns::unspecified_start;
	vstart[2] = System::Mem::AddrSpace_ns::unspecified_start;
	mf.access_IDLNS_Write = 1;
	
	aspace.map(mappables[1], region, &vstart[0], mf);
	
	region.start += Arch::page_size;
	region.end += Arch::page_size;
	aspace.map(mappables[1], region, &vstart[1], mf);

	region.start = vstart[4] + 0x12345678000ULL + kvirt_to_phys(abuf);
	region.end   = vstart[4] + 0x12345678000ULL + kvirt_to_phys(abuf) + Arch::page_size - 1;

	aspace.map(mappables[1], region, &vstart[2], mf);

	mf.CopyOnWrite = 1;
	aspace.map(mappables[1], region, &vstart[3], mf);

	printf("vstart %llx %llx %llx %llx\n", vstart[0], vstart[1], vstart[2], vstart[3]);
	char *vbuf[4];
	vbuf[0] = (char *)vstart[0];
	vbuf[1] = (char *)vstart[1];
	vbuf[2] = (char *)vstart[2];
	vbuf[3] = (char *)vstart[3];

	printf("%c%c%c%c ", vbuf[0][0], vbuf[0][1], vbuf[0][2], vbuf[0][3]);
	printf("%c%c%c%c ", vbuf[1][0], vbuf[1][1], vbuf[1][2], vbuf[1][3]);
	printf("%c%c%c%c ", vbuf[2][0], vbuf[2][1], vbuf[2][2], vbuf[2][3]);
	printf("%c%c%c%c\n", vbuf[3][0], vbuf[3][1], vbuf[3][2], vbuf[3][3]);
	
	vbuf[0][0] = 'a';
	vbuf[1][1] = 'b';
	vbuf[2][2] = 'c';
	vbuf[3][3] = 'd';
	
#if 0
	region.start = 0xdeadbeef00000000ULL + kvirt_to_phys(abuf) + Arch::page_size;
	region.end = region.start + Arch::page_size - 1;
	printf("unmapping %llx->%llx\n", region.start, region.end);

	stacked[0].unmap(region);
#endif

	printf("%c%c%c%c ", vbuf[0][0], vbuf[0][1], vbuf[0][2], vbuf[0][3]);
	printf("%c%c%c%c ", vbuf[1][0], vbuf[1][1], vbuf[1][2], vbuf[1][3]);
	printf("%c%c%c%c ", vbuf[2][0], vbuf[2][1], vbuf[2][2], vbuf[2][3]);
	printf("%c%c%c%c\n", vbuf[3][0], vbuf[3][1], vbuf[3][2], vbuf[3][3]);
}

void run_test()
{
	Threads::sched.new_thread(thread, NULL, "thread")->wake();
}
